#!/usr/bin/env python
"The rainbow server browser."

#    RN|Launcher, The rainbow server browser.
#    Copyright (C) 2008  Joakim Soderlund
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
sys.path.append("/usr/share/rnlauncher/")

import os
import gtk
import pyquake3
import urllib
import configobj

from sys import exit
from Queue import Queue
from time import sleep
from xml.dom import minidom
from threading import Thread
from SimpleGladeApp import SimpleGladeApp
from SimpleGladeApp import bindtextdomain

__author__    = "Joakim Soderlund"
__license__   = "GPLv2"
__copyright__ = "Copyright (C) 2008  Joakim Soderlund"
__date__      = "2008-08-10"
__version__   = "1.3"

cfgCompVers   = "1.0"
app_name      = "rnlauncher"
app_version   = __version__
glade_dir     = "/usr/share/rnlauncher/"
locale_dir    = ""
serverListURL = "http://rainbow.furver.se/rainbow_server_list.xml"
queryThreads  = 16

bindtextdomain(app_name, locale_dir)
gtk.gdk.threads_init()

def loadConfig():
	if os.name == "nt":
		mw.config = configobj.ConfigObj("config.ini")
	else:
		if not os.path.isdir(os.path.join(os.path.expanduser("~"), ".rnlauncher")):
			os.mkdir(os.path.join(os.path.expanduser("~"), ".rnlauncher"))
		mw.config = configobj.ConfigObj(os.path.join(os.path.expanduser("~"), ".rnlauncher", "config.ini"))

	try:
		if mw.config["main"]["cfgVer"] != cfgCompVers:
			raise Exception("Invalid config version!")
	except:
		mw.config["main"] = {"cfgVer":"1.0", "refreshOnStartup":0, "oa071bin":"", "oa077bin":""}
		mw.config.write()

	if not mw.config["main"].has_key("oa080bin"):
		mw.config["main"]["oa080bin"] = ""
		mw.config.write()

	if int(mw.config["main"]["refreshOnStartup"]):
		mw.refreshServers()

	try:
		if int(mw.config["main"]["showPlayersOnStartup"]):
			mw.mi_playersView.set_active(1)
	except:
		mw.config["main"]["showPlayersOnStartup"] = 0

	try:
		if int(mw.config["main"]["showVarsOnStartup"]):
			mw.mi_varsView.set_active(1)
	except:
			mw.config["main"]["showVarsOnStartup"] = 0

class threadedQuery(Thread):
	def run(self):
		while True:
			try:
				address, self.rowNum = mw.queue.get()
				server = pyquake3.PyQuake3(address)
				server.update()
				mw.rnServerDict[address] = server
	
				if server.vars["protocol"] == "69" and not os.path.isfile(mw.config["main"]["oa077bin"]):
						pass
				elif server.vars["protocol"] == "68" and not os.path.isfile(mw.config["main"]["oa071bin"]):
						pass
				elif server.vars["protocol"] == "70" and not os.path.isfile(mw.config["main"]["oa080bin"]):
						pass
				else:
					if server.vars["protocol"] == "69":
						gameVer = "OA 0.7.7"
					elif  server.vars["protocol"] == "68":
						gameVer = "OA 0.7.1"
					elif  server.vars["protocol"] == "70":
						gameVer = "OA 0.8.0"
					
					name = server.vars["sv_hostname"]
					name = name.split("|")
					name = name[0] + "|" + name[1]
					name = name.strip()
					name = name.strip("071")
					name = name.strip("077")
					if name.startswith("RN|^^"):
						name = name[7:]
						name = name.strip()
						name = "RN|" + name
					name = name.replace("-", "")
					name = name.replace("^", "")
					name = name.replace("*", " ")
					for a in range(10):
						name = name.replace("  ", " ")
					name = name.strip()
					if name.startswith("RN|"):
						name = name[3:]
						name = name.strip()
						name = "RN|" + name
						mw.rnServerDict[address].cleanName = name
						if self.rowNum == "all":
							mw.serverList.append([name, server.vars["mapname"], "%02d/%02d" % (len(server.players), int(server.vars["sv_maxclients"])), gameVer, address])
						else:
							mw.serverList[self.rowNum] = [name, server.vars["mapname"], "%02d/%02d" % (len(server.players), int(server.vars["sv_maxclients"])), gameVer, address]
				if hasattr(mw.queue, "task_done"):
					mw.queue.task_done()
			except:
				if hasattr(mw.queue, "task_done"):
					mw.queue.task_done()

class getServerInfo(Thread):
	"Updates the server-list."
	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)

	def run(self):
		mw.btn_refreshServers.set_sensitive(0)
		mw.btn_refreshSelected.set_sensitive(0)
		mw.btn_connect.set_sensitive(0)
		mw.btn_serverProperties.set_sensitive(0)
		mw.mi_refreshServers.set_sensitive(0)
		mw.mi_refreshSelected.set_sensitive(0)
		mw.mi_connect.set_sensitive(0)
		mw.mi_serverProperties.set_sensitive(0)
		mw.serverList.clear()
		mw.varsList.clear()
		mw.playerList.clear()
		if not os.path.isfile(mw.config["main"]["oa077bin"]) and not os.path.isfile(mw.config["main"]["oa077bin"]) and not os.path.isfile(mw.config["main"]["oa080bin"]):
			mw.statusbar.push(0, "No OpenArena binaries found! Check your settings!")
			sleep(10)
			mw.btn_refreshServers.set_sensitive(1)
			mw.mi_refreshServers.set_sensitive(1)
			mw.statusbar.push(0, " ")
			return
		mw.statusbar.push(0, "Retrieving server list from furver.se...")
		xmldata = urllib.urlopen(serverListURL).read()
		xmldoc = minidom.parseString(xmldata)
		rnServerList = []		
		for server in xmldoc.getElementsByTagName("server"):
			if server.getElementsByTagName("type")[0].firstChild.toxml() == "quake3":
				rnServerList.append(	server.getElementsByTagName("address")[0].firstChild.toxml() + ":" +
									server.getElementsByTagName("port")[0].firstChild.toxml())

		mw.statusbar.push(0, "Retrieving server info from %s servers..." % len(rnServerList))
		mw.rnServerDict = {}
		for address in rnServerList:
			mw.queue.put((address, "all"))
		if hasattr(mw.queue, "task_done"):
			mw.queue.join()
		else:
			sleep(10)
		mw.statusbar.push(0, " ")
		mw.btn_refreshServers.set_sensitive(1)
		mw.mi_refreshServers.set_sensitive(1)

class runOpenArena(Thread):
	"Launches OpenArena."
	def __init__(self, connectionString):
		Thread.__init__(self)
		self.setDaemon(True)
		self.connectionString = connectionString
	def run(self):
		os.popen(self.connectionString)

class runMessages(Thread):
	"Tells the user that we are launching OpenArena."
	def __init__(self, serverIP=None):
		Thread.__init__(self)
		self.setDaemon(True)
		if serverIP:
			self.serverName = mw.rnServerDict[serverIP].cleanName
		else:
			self.serverName = "server"
	def run(self):
		mw.btn_connect.set_sensitive(0)
		mw.btn_serverProperties.set_sensitive(0)
		mw.mi_connect.set_sensitive(0)
		mw.mi_serverProperties.set_sensitive(0)
		mw.statusbar.push(0, "Connecting to %s..." % self.serverName)
		sleep(10)
		mw.statusbar.push(0, " ")
		mw.btn_connect.set_sensitive(1)
		mw.btn_serverProperties.set_sensitive(1)
		mw.mi_connect.set_sensitive(1)
		mw.mi_serverProperties.set_sensitive(1)

# We use a thread to start the threads because we don't want
# to slow down the start of the GUI... Dirty eh?
class startQueryThreads(Thread):
	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)
	def run(self):
		for x in range(queryThreads):
			queryThread = threadedQuery()
			queryThread.setDaemon(True)
			queryThread.start()

class querySelectedServer(Thread):
	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)
	def run(self):
		mw.btn_refreshSelected.set_sensitive(0)
		mw.mi_refreshSelected.set_sensitive(0)
		serverIP, rowNum = mw.getSelectedIP(returnRowNum=True)
		mw.statusbar.push(0, "Updating %s..." % mw.rnServerDict[serverIP].cleanName)
		mw.queue.put((serverIP, rowNum))
		if hasattr(mw.queue, "task_done"):
			mw.queue.join()
		else:
			sleep(3)
		mw.updateLists(serverIP)
		mw.statusbar.push(0, "")
		mw.btn_refreshSelected.set_sensitive(1)
		mw.mi_refreshSelected.set_sensitive(1)

class mainWindow(SimpleGladeApp):
	"RN|Launcher's main window."
	def __init__(self, path="rnlauncher.glade", root="mainWindow", domain=app_name, **kwargs):
		path = os.path.join(glade_dir, path)
		SimpleGladeApp.__init__(self, path, root, domain, **kwargs)

		self.addViewColumn("Server name", 0, "serverView")
		self.addViewColumn("Map", 1, "serverView")
		self.addViewColumn("Players", 2, "serverView")
		self.addViewColumn("Version", 3, "serverView")
		self.addViewColumn("Address", 4, "serverView")

		self.serverList = gtk.ListStore(str, str, str, str, str)
		self.serverView.set_model(self.serverList)

		self.addViewColumn("Player name", 0, "playerView")
		self.addViewColumn("Score", 1, "playerView")
		self.addViewColumn("Ping", 2, "playerView")

		self.playerList = gtk.ListStore(str, str, str)
		self.playerView.set_model(self.playerList)

		self.addViewColumn("Variable", 0, "varsView")
		self.addViewColumn("Value", 1, "varsView")

		self.varsList = gtk.ListStore(str, str)
		self.varsView.set_model(self.varsList)

		self.statusbar.push(0, " ")

		self.queue = Queue()

	def addViewColumn(self, colTitle, colID, view):
		"Creats columns for the server view."
		column = gtk.TreeViewColumn(colTitle, gtk.CellRendererText(), text=colID)
		column.set_resizable(True)
		column.set_sort_column_id(colID)
		if view == "serverView":
			self.serverView.append_column(column)
		elif view == "playerView":
			self.playerView.append_column(column)
		elif view == "varsView":
			self.varsView.append_column(column)
		
	def on_btn_refreshServers_clicked(self, widget, *args):
		"Called when refresh servers is clicked."
		self.refreshServers()

	def on_btn_refreshSelected_clicked(self, widget, *args):
		"Called when refresh selected is clicked."
		self.querySelectedServer = querySelectedServer()
		self.querySelectedServer.start()

	def refreshServers(self):
		"Refreshes the server list."
		self.getServerInfo = getServerInfo()
		self.getServerInfo.start()

	def getSelectedIP(self, returnRowNum=False):
		"Returns the address of the selected server."
		rowNum = int("%s" % self.serverView.get_selection().get_selected_rows()[1][0])
		item = self.serverList.get_iter_first()
		cats = []
		while ( item != None ):
			cats.append(self.serverList.get_value (item, 4))
			item = self.serverList.iter_next(item)
		if returnRowNum:
			return cats[rowNum], rowNum
		else:
			return cats[rowNum]

	def connectToServer(self):
		"Connects to the selected server."
		serverIP = self.getSelectedIP()

		if self.rnServerDict[serverIP].vars["protocol"] == "69":
			oabin = self.config["main"]["oa077bin"]
		elif self.rnServerDict[serverIP].vars["protocol"] == "68":
			oabin = self.config["main"]["oa071bin"]
		elif self.rnServerDict[serverIP].vars["protocol"] == "70":
			oabin = self.config["main"]["oa080bin"]
		else:
			raise Exception("Error! Launched incompatible protocol version!")

		try:
			rconpassword = self.config[serverIP]["rconpassword"]
		except:
			rconpassword = ""
		self.runOpenArena = runOpenArena("%s +connect \"%s\" +set rconpassword \"%s\"" % (oabin, serverIP, rconpassword))
		self.runMessages = runMessages(serverIP)
		self.runMessages.start()
		self.runOpenArena.start()


	def on_btn_connect_clicked(self, widget, *args):
		self.connectToServer()

	def on_btn_serverProperties_clicked(self, widget, *args):
		serverIP = self.getSelectedIP()
		self.serverProperties = serverSettings()
		try:
			rconpassword = self.config[serverIP]["rconpassword"]
		except:
			rconpassword = ""
		self.serverProperties.recieveServerInfo(serverIP, self.rnServerDict[serverIP], rconpassword)
		self.serverProperties.run()

	def on_btn_mainSettings_clicked(self, widget, *args):
		self.mainSettings = mainSettings()
		self.mainSettings.run()

	def on_serverView_row_activated(self, widget, *args):
		self.connectToServer()

	def on_serverView_cursor_changed(self, widget, *args):
		"Called when a server has been selected."
		mw.btn_refreshSelected.set_sensitive(1)
		mw.btn_connect.set_sensitive(1)
		mw.btn_serverProperties.set_sensitive(1)
		mw.mi_refreshSelected.set_sensitive(1)
		mw.mi_connect.set_sensitive(1)
		mw.mi_serverProperties.set_sensitive(1)
		self.updateLists(self.getSelectedIP())

	def updateLists(self, serverIP):
		"Updates the player list and vars list."
		self.playerList.clear()
		self.varsList.clear()
		for player in self.rnServerDict[serverIP].players:
			player.name_clean = player.name
			# Removes the colors from the player colors
			for integer in range(9):
				player.name_clean = player.name_clean.replace("^%d" % integer, "")
			self.playerList.append([player.name_clean, "%03d" % int(player.frags), "%03d" % int(player.ping)])
		for k, v in self.rnServerDict[serverIP].vars.items():
			self.varsList.append([k, v])

	def on_mi_playersView_toggled(self, widget, *args):
		"Shows or hides the player view."
		if self.mi_playersView.get_active():
			self.playerViewWindow.show()
		else:
			self.playerViewWindow.hide()

	def on_mi_varsView_toggled(self, widget, *args):
		"Shows or hides the vars view."
		if self.mi_varsView.get_active():
			self.varsViewWindow.show()
		else:
			self.varsViewWindow.hide()

	def on_menu_about_activate(self, widget, *args):
		"Shows the about dialog."
		aboutdialog = aboutDialog()
		aboutdialog.run()

	def on_mainWindow_destroy(self, widget, *args):
		"Destroys the main window."
		gtk.main_quit()

	def run(self):
		"Initializes GTK."
		gtk.gdk.threads_enter()
		gtk.main()
		gtk.gdk.threads_leave()

class aboutDialog(SimpleGladeApp):
	def __init__(self, path="rnlauncher.glade", root="aboutDialog", domain=app_name, **kwargs):
		path = os.path.join(glade_dir, path)
		SimpleGladeApp.__init__(self, path, root, domain, **kwargs)
		self.aboutDialog.set_version(__version__)

	def on_aboutDialog_destroy(self, widget, *args):
		self.aboutDialog.destroy()
		gtk.main_quit()

class mainSettings(SimpleGladeApp):
	def __init__(self, path="rnlauncher.glade", root="mainSettings", domain=app_name, **kwargs):
		path = os.path.join(glade_dir, path)
		SimpleGladeApp.__init__(self, path, root, domain, **kwargs)

		self.ent_oa071bin.set_text(mw.config["main"]["oa071bin"])
		self.ent_oa077bin.set_text(mw.config["main"]["oa077bin"])
		self.ent_oa080bin.set_text(mw.config["main"]["oa080bin"])
		self.cbx_refreshOnStartup.set_active(int(mw.config["main"]["refreshOnStartup"]))
		self.cbx_showPlayersOnStartup.set_active(int(mw.config["main"]["showPlayersOnStartup"]))
		self.cbx_showVarsOnStartup.set_active(int(mw.config["main"]["showVarsOnStartup"]))

	def on_btn_ok2_clicked(self, widget, *args):
		mw.config["main"]["oa071bin"] = self.ent_oa071bin.get_text()
		mw.config["main"]["oa077bin"] = self.ent_oa077bin.get_text()
		mw.config["main"]["oa080bin"] = self.ent_oa080bin.get_text()
		if self.cbx_refreshOnStartup.get_active():
			mw.config["main"]["refreshOnStartup"] = 1
		else:
			mw.config["main"]["refreshOnStartup"] = 0

		if self.cbx_showPlayersOnStartup.get_active():
			mw.config["main"]["showPlayersOnStartup"] = 1
		else:
			mw.config["main"]["showPlayersOnStartup"] = 0

		if self.cbx_showVarsOnStartup.get_active():
			mw.config["main"]["showVarsOnStartup"] = 1
		else:
			mw.config["main"]["showVarsOnStartup"] = 0

		mw.config.write()
		self.mainSettings.destroy()

	def on_btn_oa071bin_clicked(self, widget, *args):
		self.fileChooser = fileChooser()
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.7.1 binary")
		self.fileChooser.run()
		try:
			self.ent_oa071bin.set_text(self.fileChooser.filename)
		except AttributeError:
			pass

	def on_btn_oa077bin_clicked(self, widget, *args):
		self.fileChooser = fileChooser()
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.7.7 binary")
		self.fileChooser.run()
		try:
			self.ent_oa077bin.set_text(self.fileChooser.filename)
		except AttributeError:
			pass

	def on_btn_oa080bin_clicked(self, widget, *args):
		self.fileChooser = fileChooser()
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.8.0 binary")
		self.fileChooser.run()
		try:
			self.ent_oa080bin.set_text(self.fileChooser.filename)
		except AttributeError:
			pass

	def on_btn_abort2_clicked(self, widget, *args):
		self.mainSettings.destroy()

	def on_mainSettings_destroy(self, widget, *args):
		gtk.main_quit()



class serverSettings(SimpleGladeApp):
	def __init__(self, path="rnlauncher.glade", root="serverSettings", domain=app_name, **kwargs):
		path = os.path.join(glade_dir, path)
		SimpleGladeApp.__init__(self, path, root, domain, **kwargs)


	def recieveServerInfo(self, serverIP, serverData, rconpassword):
		self.serverIP = serverIP
		self.serverData = serverData
		self.lab_serverName.set_text(serverIP)
		self.ent_rconPassword.set_text(rconpassword)

	def on_btn_ok_clicked(self, widget, *args):
		try:
			mw.config[self.serverIP]["rconpassword"] = self.ent_rconPassword.get_text()
		except KeyError:
			mw.config[self.serverIP] = {}
			mw.config[self.serverIP]["rconpassword"] = self.ent_rconPassword.get_text()
		mw.config.write()
		self.serverSettings.destroy()
		
	def on_btn_abort_clicked(self, widget, *args):
		self.serverSettings.destroy()

	def on_serverSettings_destroy(self, widget, *args):
		gtk.main_quit()

class fileChooser(SimpleGladeApp):
	def __init__(self, path="rnlauncher.glade", root="fileChooser", domain=app_name, **kwargs):
		path = os.path.join(glade_dir, path)
		SimpleGladeApp.__init__(self, path, root, domain, **kwargs)

	def on_btn_okfc_clicked(self, widget, *args):
		self.filename = self.fileChooser.get_filename()
		self.fileChooser.destroy()
		gtk.main_quit()

	def on_btn_abortfc_clicked(self, widget, *args):
		self.fileChooser.destroy()
		gtk.main_quit()

	def on_fileChooser_destroy(self, widget, *args):
		gtk.main_quit()

if __name__ == "__main__":
	try:
		mw = mainWindow()
		loadConfig()
		mw.startQueryThreads = startQueryThreads()
		mw.startQueryThreads.start()
		mw.run()
		mw.config.write()
	except:
		pass
	exit()
